# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
from datetime import datetime
from sqlalchemy_aio import ASYNCIO_STRATEGY
from sqlalchemy import create_engine, MetaData, Table, Column,\
                       Integer, String, Text, Boolean, DateTime, create_engine
from sqlalchemy.schema import CreateTable
from sqlalchemy.sql import select
from livebridge.storages.base import BaseStorage


logger = logging.getLogger(__name__)

"""
http://docs.sqlalchemy.org/en/latest/core/engines.html
"""
class SQLStorage(BaseStorage):
    
    _instance = None
    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(SQLStorage, cls).__new__(cls)
            logger.debug("SQL client: {}".format(cls._instance))
        return cls._instance

    def __init__(self, **kwargs):
        self.dsn = kwargs.get("dsn", None)
        self.table_name = kwargs.get("table_name")
        
    @property
    async def db(self):
        if hasattr(self, "_engine") and self._engine:
            return self._engine
        logger.debug("Connecting to {}".format(self.dsn))
        self._engine = create_engine(self.dsn, strategy=ASYNCIO_STRATEGY)
        return self._engine

    def _get_table(self):
        return Table(self.table_name, MetaData(),
            Column("id", Integer(), primary_key=True),
            Column("target_id", String(150), index=True),
            Column("post_id", String(150), index=True),
            Column("source_id", String(150), index=True),
            Column("sticky", Boolean()),
            Column("text", Text()),
            Column("created", DateTime()),
            Column("updated", DateTime(), index=True),
            Column("target_doc", Text()),
        )

    async def setup(self):
        """Setting up SQL table, if it not exists."""
        try:
            engine = await self.db
            if not await engine.has_table(self.table_name):
                # create table
                logger.info("Creating SQL table [{}]".format(self.table_name))
                items = self._get_table()
                await engine.execute(CreateTable(items))
                # create indeces
                conn = await engine.connect()
                await conn.execute("CREATE INDEX `lb_last_updated` ON `{}` (`source_id` DESC,`updated` DESC);".format(self.table_name))
                await conn.execute("CREATE INDEX `lb_post` ON `{}` (`target_id` DESC,`post_id` DESC);".format(self.table_name))
                await conn.close()
                return True
        except Exception as e:
            logger.error("[DB] Error when setting up SQL table: {}".format(e))
        return False

    async def get_last_updated(self, source_id):
        try:
            db = await self.db
            table = self._get_table()
            result = await db.execute(table.select().where(table.c.source_id==source_id).limit(1))
            item = await result.first()
            tstamp = item["updated"] if item else datetime.utcnow()
            return tstamp
        except Exception as e:
            logger.error("[DB] Error when querying for last updated item on {}".format(source_id))
            logger.exception(e)
        return None

    async def get_known_posts(self, source_id, post_ids):
        results = []
        try:
            db = await self.db
            table = self._get_table()
            sql = select(columns=[table.c.post_id]).where(table.c.source_id==source_id).where(table.c.post_id.in_(post_ids))
            db_res = await db.execute(sql)
            for r in await db_res.fetchall():
                results.append(r[0])
        except Exception as e:
            logger.error("[DB] Error when querying for posts {}".format(post_ids))
            logger.exception(e)
        return results

    async def get_post(self, target_id, post_id):
        try:
            db = await self.db
            table = self._get_table()
            sql = table.select().where(table.c.target_id==target_id).where(table.c.post_id==post_id).limit(1)
            result = await db.execute(sql)
            item = await result.first()
            if item:
                item = dict(item)
                item["target_doc"] = json.loads(item["target_doc"]) if item.get("target_doc") != "" else {}
            return item
        except Exception as e:
            logger.error("[DB] Error when querying for a post [{}] on {}".format(post_id, target_id))
            logger.error(e)
        return None

    async def insert_post(self, **kwargs):
        try:
            db = await self.db
            table = self._get_table()
            sql = table.insert().values(
                target_id=kwargs.get("target_id"), 
                post_id=kwargs.get("post_id"), 
                source_id=kwargs.get("source_id"), 
                text=kwargs.get("text"), 
                sticky=int(kwargs.get("sticky", "0")), 
                created=kwargs.get("created"),
                updated=kwargs.get("updated"),
                target_doc=json.dumps(kwargs.get("target_doc")) if kwargs.get("target_doc") else "" 
            ) 
            await db.execute(sql)
            logger.info("[DB] Post {} {} was saved!".format(kwargs["source_id"], kwargs["post_id"]))
            return True
        except Exception as e:
            logger.error("[DB] Error when saving {}".format(kwargs))
            logger.error(e)
        return False

    async def update_post(self, **kwargs):
        try:
            db = await self.db
            table = self._get_table()
            sql = table.update().where(
                    table.c.target_id==kwargs.get("target_id")
                ).where(
                    table.c.post_id==kwargs.get("post_id")
                ).values(
                    target_id=kwargs.get("target_id"), 
                    post_id=kwargs.get("post_id"), 
                    source_id=kwargs.get("source_id"), 
                    text=kwargs.get("text"), 
                    sticky=int(kwargs.get("sticky", "0")), 
                    created=kwargs.get("created"),
                    updated=kwargs.get("updated"),
                    target_doc=json.dumps(kwargs.get("target_doc")) if kwargs.get("target_doc") else "" 
                ) 
            await db.execute(sql)
            logger.info("[DB] Post {} {} was updated!".format(kwargs.get("post_id"), kwargs.get("target_id")))
            return True
        except Exception as e:
            logger.error("[DB] Error when updating for a post [{}] on {}".format(kwargs.get("post_id"), kwargs.get("target_id")))
            logger.error(e)
        return False

    async def delete_post(self, target_id, post_id):
        try:
            db = await self.db
            table = self._get_table()
            sql = table.delete().where(table.c.target_id==target_id).where(table.c.post_id==post_id)
            await db.execute(sql)
            logger.info("[DB] Post {} {} was deleted!".format(target_id, post_id))
            return True
        except Exception as e:
            logger.error("[DB] Error when deleting for a post [{}] on {}".format(post_id, target_id))
            logger.error(e)
        return False



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import boto3
import logging
import yaml
from botocore.client import Config
from livebridge.config import AWS

logger = logging.getLogger(__name__)

class ControlFile(object):

    def __init__(self):
        pass

    def _resolve_auth(self, data):
        for x, bridge in enumerate(data.get("bridges", [])):
            if data.get("auth", {}).get(bridge.get("auth")):
                # add user creds to bridge config
                data["bridges"][x]["auth"] = data["auth"][bridge["auth"]]
            for y, target in enumerate(bridge.get("targets", [])):
                if data.get("auth", {}).get(target.get("auth")):
                    # add user creds to target config
                    data["bridges"][x]["targets"][y]["auth"] = data["auth"][target["auth"]]
                elif target.get("auth"):
                    raise LookupError("Data for user [{}] not found in control file.".format(target.get("auth")))
        return data

    def load(self, path, *, resolve_auth=False):
        if not path.startswith("s3://"):
            body = self.load_from_file(path)
        else:
            body = self.load_from_s3(path)
        control = yaml.load(body)

        if resolve_auth:
            control= self._resolve_auth(control)

        return control

    def load_from_file(self, path):
        logger.info("Loading control file from disk: {}".format(path))
        if not os.path.exists(path):
            raise IOError("Path for control file not found.")

        f = open(path, "r")
        body = f.read()
        f.close()

        return body

    def load_from_s3(self, url):
        bucket, key = url.split('/',2)[-1].split('/',1)
        logger.info("Loading control file from s3: {} - {}".format(bucket, key))
        session = boto3.session.Session()
        config = Config(signature_version="s3v4") if AWS["region"] in ["eu-central-1"] else None
        client = boto3.client(
            's3',
            region_name=AWS["region"],
            aws_access_key_id=AWS["access_key"] or None,
            aws_secret_access_key=AWS["secret_key"] or None,
            config=config,
        )
        control_file = client.get_object(Bucket=bucket, Key=key)
        return control_file["Body"].read()



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import aiobotocore
import asyncio
import logging
import json
from botocore.exceptions import ClientError
from livebridge.components import get_target
from livebridge.bridge import LiveBridge
from livebridge.controlfile import ControlFile

logger = logging.getLogger(__name__)

class Controller(object):

    def __init__(self, config, control_file, poll_interval=60):
        self.config = config
        self.control_file = control_file
        self.poll_interval = poll_interval
        self.read_control = False
        self.tasked = []
        self.sleep_tasks = []
        self.bridges = {}
        self._sqs_client = None
        self.retry_run_interval = 30
        self.control_data = None # data from control file
        self.shutdown = False

    def __del__(self):
        if self._sqs_client:
            self._sqs_client.close()

    @property
    def sqs_client(self):
        if self._sqs_client:
            return self._sqs_client
        loop = asyncio.get_event_loop()
        session = aiobotocore.get_session(loop=loop)
        self._sqs_client = session.create_client('sqs',
                                    region_name = self.config["region"],
                                    aws_secret_access_key = self.config["secret_key"] or None,
                                    aws_access_key_id = self.config["access_key"] or None)
        return self._sqs_client

    async def clean_shutdown(self):
        logger.info("Requesting proper shutdown of tasks.")
        self.shutdown = True
        await self.stop_bridges()
        while len(self.bridges) > 0:
            logger.debug("Running bridges left: {}".format(len(self.bridges)))
            await asyncio.sleep(1)

    async def stop_bridges(self):
        """Stop all sleep tasks to allow bridges to end."""
        for t in self.sleep_tasks:
            t.cancel()

    async def check_control_change(self):
        client = self.sqs_client
        logger.info("Starting watching for control file changes on s3.")
        try:
            # purge queue before starting watching
            purge_resp = await client.purge_queue(
                QueueUrl=self.config["sqs_s3_queue"]
            )
            logger.info("Purged SQS queue {}".format(self.config["sqs_s3_queue"]))
        except ClientError as e:
            logger.warning("Purging SQS queue failed with: {}".format(e))
        # check for update events
        while True and self.shutdown != True:
            try:
                response = await client.receive_message(
                    QueueUrl=self.config["sqs_s3_queue"]
                )
                for msg in response.get("Messages", []):
                    logger.debug("SQS {}".format(msg.get("MessageId")))
                    body = msg.get("Body")
                    data = json.loads(body) if body else None
                    del_res = await client.delete_message(
                        QueueUrl=self.config["sqs_s3_queue"],
                        ReceiptHandle=msg.get("ReceiptHandle")
                    )
                    if data:
                        for rec in data.get("Records", []):
                            logger.debug("EVENT: {} {}".format(rec.get("s3", {}).get("object", {}).get("key"), rec.get("eventName")))
                            self.read_control = True
                            await self.stop_bridges()
                            return
            except Exception as e:
                logger.error("Error fetching SQS messages with: {}".format(e))
            await self.sleep(60)
        client.close()

    def append_bridge(self, config_data):
        bridge = LiveBridge(config_data)
        for t in config_data.get("targets", []):
            target_client = get_target(t)
            bridge.add_target(target_client)
        # start specific source 
        if bridge.source.mode == "streaming":
            self.bridges[bridge] = self.run_stream(bridge=bridge)
        elif bridge.source.mode == "polling":
            # custom source polling intervall set in control file?
            poll_interval = config_data["poll_interval"] if config_data.get("poll_interval") else self.poll_interval
            self.bridges[bridge] = self.run_poller(bridge=bridge, interval=poll_interval)

    def remove_bridge(self, bridge):
        logger.info("ENDING: {}".format(bridge))
        task = self.bridges[bridge]
        del self.bridges[bridge]
        if self.read_control and not self.bridges:
            logger.info("RESTART BRIDGES")
            self.run()

    async def retry_run(self):
        logger.info("Will retry loading control file in 30 seconds.")
        await asyncio.sleep(self.retry_run_interval)
        self.run()

    def run(self):
        """ Blocking code."""
        self.read_control = False
        control_data = None
        try:
            c = ControlFile()
            control_data = c.load(self.control_file, resolve_auth=True)
        except Exception as e:
            logger.error("Error when reading control file.")
            logger.error(e)
            logger.info("Will try to reuse existing control data.")

        # set new control data
        if control_data:
            logger.info("Using fetched control data.")
            self.control_data = control_data

        # (re)start tasks or retry fetching control file
        if self.control_data:
            self.start_tasks()
        else:
            self.tasked.append(asyncio.Task(self.retry_run()))

    
    def start_tasks(self):
        # start content bridges
        for b in self.control_data["bridges"]:
            self.append_bridge(b)

        # create futures for bridges
        for b in self.bridges:
            self.tasked.append(asyncio.Task(self.bridges[b]))

        # listen to s3 control changes if sqs queue is given
        if self.config.get("sqs_s3_queue"):
            self.tasked.append(asyncio.Task(self.check_control_change()))

    async def run_stream(self, *, bridge):
        future = await bridge.listen_ws()
        while True and self.read_control != True and self.shutdown != True:
            # wait for shutdown
            await self.sleep(4)

        # stop bridge
        try:
            await bridge.source.stop()
        except Exception as e:
            logger.error("Error when stopping stream: {}".format(e))

        self.remove_bridge(bridge)
        return 

    async def sleep(self, seconds):
        if self.read_control == True or self.shutdown == True:
            # if shutdown or restart is requested, don't fall asleep again
            return True

        try:
            task = asyncio.ensure_future(asyncio.sleep(seconds))
            self.sleep_tasks.append(task)
            await task
        except asyncio.CancelledError:
            pass
        finally:
            self.sleep_tasks.remove(task)
        return True

    async def run_poller(self, *, bridge, interval=180):
        # initialize liveblogs to watch
        while True and self.read_control != True and self.shutdown != True:
            #logger.debug("Checked new posts for {} on {}".format(bridge.source_id, bridge.endpoint))
            await bridge.check_posts()
            await self.sleep(interval)

        self.remove_bridge(bridge)
        return 



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class BaseStorage(object):

    @property
    async def db(self):
        raise NotImplementedError()

    def setup(self):
        raise NotImplementedError()

    async def insert_post(self, **kwargs):
        raise NotImplementedError()

    async def get_post(self, target_id, post_id):
        raise NotImplementedError()

    async def update_post(self, target_id, post_id, ):
        raise NotImplementedError()

    async def delete_post(self, target_id, post_id):
        raise NotImplementedError()

    async def get_last_updated(self, source_id):
        raise NotImplementedError()

    async def get_known_posts(self, source_id, post_ids):
        """Return a list of known post_id of a source for a given list of post ids.

        :param source_id: id of the source
        :type string:
        :param post_ids: list of post ids
        :type list:
        :returns: - list of dictionaries."""
        raise NotImplementedError()



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import copy
import logging
from livebridge.components import get_converter, get_post, get_source, get_db_client


logger = logging.getLogger(__name__)


class LiveBridge(object):

    def __init__(self, config):
        self.targets = []
        self.api_client = None
        self.config = config
        self.source_id = self.config.get("source_id")
        self.endpoint = self.config.get("endpoint")
        self.label = self.config.get("label", "-")
        self.db = get_db_client()

    def __repr__(self):
        return "<LiveBridge [{}] {} {}>".format(self.label, self.endpoint, self.source_id)

    @property
    def source(self):
        if self.api_client:
            return self.api_client
        # setup api client
        self.api_client = get_source(self.config)
        return self.api_client

    def add_target(self, target):
        logger.debug("Adding target {} to {}".format(target, self))
        self.targets.append(target)

    async def listen_ws(self):
        return asyncio.Task(self.source.listen(self.new_posts))

    async def check_posts(self):
        try:
            # set last_updated
            if not self.source.last_updated:
                self.source.last_updated = await self.db.get_last_updated(self.source_id)
                logger.info("LAST UPDATED: {} {}".format(self.source.last_updated, self.source))
            posts = await self.source.poll()
            if posts:
                await self.new_posts(posts)
        except Exception as e:
            logger.error("Fatal checking {}{}".format(getattr(self, "endpoint", "-"), self.source_id))
            logger.exception(e)
        return True

    async def _process_post(self, target, source_post):
        post = copy.deepcopy(source_post)
        # update last updated property in source client
        self.source.last_updated = post.updated
        # handle post in target
        await target.handle_post(post)

    async def new_posts(self, posts):
        logger.info("##### Received {} posts from {}".format(len(posts), self.source))
        for p in posts:
            coros = [self._process_post(t, p) for t in self.targets]
            await asyncio.gather(*coros)
        return True
